//
//  DiaryPrintView.m
//  Vermont Recipes 2.0.0
//
//  Created by Bill Cheeseman on 2009-11-08.
//  Copyright 2000-2009 Bill Cheeseman. All rights reserved.
//
//  CREATED IN RECIPE 9

#import "DiaryPrintView.h"
#import "DiaryWindowController.h" // ADDED IN RECIPE 9
#import "DiaryPrintPanelAccessoryController.h" // ADDED IN RECIPE 9


#pragma mark MACROS

#define DEFAULT_ALERT_PRINT_SCALED_UP_SUPPRESSED_KEY	@"alert print scaled up suppressed" // ADDED IN RECIPE 9

#pragma mark -

@implementation DiaryPrintView

#pragma mark INITIALIZATION	

- (id)initWithFrame:(NSRect)frame { // ADDED IN RECIPE 9
    self = [super initWithFrame:frame];
    if (self) {
        // Initialization code here.
		// The frame argument is passed in as NSZeroRect from -[DiaryDocument printOperationWithSettings:error:. The print operation has not yet run, so the values in the document's print info dictionary are not yet available here. The frame is set in -knowsPageRange after the print operation has run.
		currentDocument = [[[NSDocumentController sharedDocumentController] currentDocument] retain];
		currentDate = [[NSDate date] retain];
    }
    return self;
}

- (void)dealloc { // ADDED IN RECIPE 9
	[textStorage release]; // instantiated in -knowsPageRange:
	[currentOperation release]; // instantiated in -knowsPageRange:
	[currentDate release];
	[currentDocument release];
	[super dealloc];
}

#pragma mark ACCESSOR METHODS

- (void)setCurrentDocument:(DiaryDocument *)document{ // ADDED IN RECIPE 9
	if (currentDocument != document) {
		[currentDocument release];
		currentDocument = [document retain];
	}
}

- (DiaryDocument *)currentDocument { // ADDED IN RECIPE 9
	return [[currentDocument retain] autorelease];
}

- (void)setCurrentDate:(NSDate *)date { // ADDED IN RECIPE 9
	if (currentDate != date) {
		[currentDate release];
		currentDate = [date retain];
	}
}

- (NSDate *)currentDate { // ADDED IN RECIPE 9
	return [[currentDate retain] autorelease];
}

- (void)setCurrentOperation:(NSPrintOperation *)printOperation { // ADDED IN RECIPE 9
	if (currentOperation != printOperation) {
		[currentOperation release];
		currentOperation = [printOperation retain];
	}
}

- (NSPrintOperation *)currentOperation { // ADDED IN RECIPE 9
	return [[currentOperation retain] autorelease];
}

- (void)setTextStorage:(NSTextStorage *)storage { // ADDED IN RECIPE 9
	if (textStorage != storage) {
		[textStorage release];
		textStorage = [storage retain];
	}
}

- (NSTextStorage *)textStorage { // ADDED IN RECIPE 9
	return [[textStorage retain] autorelease];
}

#pragma mark OVERRIDE METHODS

// These methods are called in the order given, when the print panel opens, every time the user changes settings in the print panel, and when the document is printed. 

- (BOOL)isFlipped { // ADDED IN RECIPE 9
	// The glyph-based routines in -paginateContent and -drawRect: require a flipped view (origin at top-left corner).
	return YES;
}

- (BOOL)knowsPageRange:(NSRangePointer)range { // ADDED IN RECIPE 9
	// This is the first method called by the Cocoa printing system for every print session. This method is called for the entire document. Construct and paginate the entire printable content of the document here, returning its page range indirectly in the range argument and returning YES because the diary document supports custom pagination. Recalculates page height, page width, printable content, and page count, and resets the frame, because the user may change paper size, margins, and print items at any time. At this point the print operation has been run, so this is the first opportunity to get and set the values in this view's copy of the document's print info dictionary. You must set values in the print info dictionary that affect pagination here. You can optionally set and get values that don't affect pagination either here or in -beginDocument.
	[self setCurrentOperation:[NSPrintOperation currentOperation]];
	NSPrintInfo *printInfo = [[self currentOperation] printInfo];

	// Prevent scaling at more than 100%.
	CGFloat scalingFactor;
	if (floor(NSAppKitVersionNumber) <= NSAppKitVersionNumber10_5) {
		scalingFactor = [[[printInfo dictionary] objectForKey:NSPrintScalingFactor] floatValue];
		if (scalingFactor > 1.0) {
			[[printInfo dictionary] setObject:[NSNumber numberWithFloat:1.0] forKey:NSPrintScalingFactor];
		}
	} else {
		scalingFactor = [printInfo scalingFactor];
		if (scalingFactor > 1.0) {
			[printInfo setScalingFactor:1.0];
		}
	}
	
	// Present warning alert
	if (scalingFactor > 1.0) {
		NSUserDefaults *defaults = [NSUserDefaults standardUserDefaults];
		if (![defaults boolForKey:DEFAULT_ALERT_PRINT_SCALED_UP_SUPPRESSED_KEY]) {
			NSAlert *alert = [self alertCannotPrintScaledUp];
			[alert runModal];
			[defaults setBool:[[alert suppressionButton] state] forKey:DEFAULT_ALERT_PRINT_SCALED_UP_SUPPRESSED_KEY];
			[defaults synchronize];
		}
	}
	
	// Calculate size of content area of printed page based on paper size and margin settings.
	NSSize paperSize = [printInfo paperSize];
	NSSize pageSize = NSMakeSize(paperSize.width - [printInfo leftMargin] - [printInfo rightMargin], paperSize.height - [printInfo topMargin] - [printInfo bottomMargin]);
	
	// Construct printable content based on settings in the print panel (e.g., printItems and printTags). We create a Cocoa text system text storage object to hold the printable content.
	[self setTextStorage:[self printableContent]];
	
	// Set the frame to an arbitrarily large height.
	[self setFrame:NSMakeRect(0.0, 0.0, pageSize.width, 1.0e7)];
	// NSRect boundingRect = [[self textStorage] boundingRectWithSize:NSMakeSize(pageSize.width, 0.0) options:NSStringDrawingUsesLineFragmentOrigin | NSStringDrawingUsesFontLeading | NSStringDrawingDisableScreenFontSubstitution]; // 0.0 signals height is flexible
	// [self setFrame:boundingRect];
	
	// Paginate printable content and calculate page count. We create a Cocoa text system text container object for each page.
	[self paginateContent:pageSize];
	NSUInteger pageCount = [[[[[self textStorage] layoutManagers] objectAtIndex:0] textContainers] count];
	
	// Return range values indirectly to the Cocoa text system.
	range->location = 1;
	range->length = pageCount;
	
	// Tell the Cocoa text system we have paginated the entire printable content of the document.
	return YES;
}

- (void)beginDocument { // ADDED IN RECIPE 9
	// Called by the Cocoa printing system after -knowsPageRange and before -rectForPage. This method is called for the entire document. You can optionally set and get values in the document's print info dictionary that don't affect pagination here instead of in -knowsPageRange. You must call super because it sets up the current graphics context. If you have no code here, delete this override method.
	[super beginDocument];
	
	// Set printInfo properties.
	NSPrintInfo *printInfo = [[self currentOperation] printInfo];
	[printInfo setHorizontallyCentered:NO];
	[printInfo setVerticallyCentered:NO];
	
	// Set print info's NSPrintHeaderAndFooter key to turn page headers and footers on and off. We need this only because we use our own VRDefaultDiaryPrintHeadersAndFootersKey key to capture this print panel setting.
	NSNumber *doPrintHeadersAndFooters = [[printInfo dictionary] objectForKey:VRDefaultDiaryPrintHeadersAndFootersKey];
	[[printInfo dictionary] setObject:doPrintHeadersAndFooters forKey:NSPrintHeaderAndFooter];
}

- (NSRect)rectForPage:(NSInteger)page { // ADDED IN RECIPE 9
	// Called by the Cocoa printing system after -beginDocument and before -beginPageInRect, and any time you need to know the actual printing rect of the current page. This method is called for every page. Calculate and return the rect of the current page within the coordinate system of the entire view, including the page's origin based on the cumulative height of previous pages, if any. Because this is a text document, you have set the view to be flipped, with the origin at the top left. Returns the rect for the page content excluding margins. The page argument is a one-based index. You must set values that affect this page's rect here. You can optionally set and get values that don't affect this page's rect either here or in -beginPageInRect.
	NSLayoutManager *layoutManager = [[[self textStorage] layoutManagers] objectAtIndex:0];
	NSArray *textContainerArray = [layoutManager textContainers];
	NSTextContainer *thisTextContainer;
	CGFloat pageOriginY = 0.0;
	for (NSUInteger idx = 0; idx < page; idx++) {
		thisTextContainer = [textContainerArray objectAtIndex:idx];
		if (idx < page - 1) pageOriginY += [thisTextContainer containerSize].height;
	}
	return NSMakeRect(0.0, pageOriginY, [thisTextContainer containerSize].width, [thisTextContainer containerSize].height);
}

- (void)beginPageInRect:(NSRect)rect atPlacement:(NSPoint)location { // ADDED IN RECIPE 9
	// Called by the Cocoa printing system after -rectForPage and before -drawPageBorderWithSize. This method is called for every page. You can optionally set and get values in the document's print info dictionary that don't affect this page's rect here instead of in -rectForPage:. The one-based page number is available by calling [[self currentOperation] currentPage]. You must call super because it sets up the coordinate system and performs translation. If you have no code here, delete this override method.
	[super beginPageInRect:rect atPlacement:location];

	// Scale the page content. Header and footer scaling is handled by -drawPageBorderWithSize:.
	NSPrintInfo *printInfo = [[self currentOperation] printInfo];
	CGFloat scalingFactor;
	if (floor(NSAppKitVersionNumber) <= NSAppKitVersionNumber10_5) {
		scalingFactor = [[[printInfo dictionary] objectForKey:NSPrintScalingFactor] floatValue];
	} else {
		scalingFactor = [printInfo scalingFactor];
	}
	NSAffineTransform *transform = [NSAffineTransform transform];
	[transform scaleBy:scalingFactor];
	[transform concat];
}

- (void)drawPageBorderWithSize:(NSSize)borderSize { // ADDED IN RECIPE 9
	// Called by the Cocoa printing system after -beginPageInRect:atLocation: and before -drawRect:. This method is called for every page. Use this method to print in the margins or anywhere within the paper's rect using -drawAtPoint: or similar methods, even overprinting the page content, such as when printing watermarks. By default, borderSize is equal to paperSize. The page's origin within this method is the bottom left corner of the page, and the rect for the sheet of paper is {0.0, 0.0, borderSize.width, borderSize.height}, even though the DiaryPrintView is flipped. It is not necessary to adjust the view's frame. Be sure to pay attention to the rect returned by -[NSPrintInfo imageablePageBounds], which varies among printers; it is typically inset from the paper's rect on all four edges unless the selected paper size is borderless. Marks printed at the edge of the paper when printing borderless may actually be placed beyond the edge of the paper due to discrepancies unique to the printer; you should position and size the marks to accommodate this.
	// The Cocoa printing system automatically scales everything drawn in this method to the origin at the bottom left corner of the page, including the margins. Here, we reposition the headers and footers to the top left corner and honor the unscaled margins.
	NSPrintInfo *printInfo = [[self currentOperation] printInfo];
	CGFloat scalingFactor;
	if (floor(NSAppKitVersionNumber) <= NSAppKitVersionNumber10_5) {
		scalingFactor = [[[printInfo dictionary] objectForKey:NSPrintScalingFactor] floatValue];
	} else {
		scalingFactor = [printInfo scalingFactor];
	}
	
	// Print paper corner marks for demonstration purposes. These will not print if they are outside the current printer's imageable page bounds, unless you are printing to PDF or printing borderless.
	if (scalingFactor == 1.0) { // print only if not scaled
		NSString *cornerMark = @"XX\nXX"; // 4 X's in a square arrangement
		[cornerMark drawAtPoint:NSZeroPoint withAttributes:nil];
		[cornerMark drawAtPoint:NSMakePoint(0.0, borderSize.height - [cornerMark sizeWithAttributes:nil].height) withAttributes:nil];
		[cornerMark drawAtPoint:NSMakePoint(borderSize.width - [cornerMark sizeWithAttributes:nil].width, 0.0) withAttributes:nil];
		[cornerMark drawAtPoint:NSMakePoint(borderSize.width - [cornerMark sizeWithAttributes:nil].width, borderSize.height - [cornerMark sizeWithAttributes:nil].height) withAttributes:nil];
	}
	
	// Print a custom header and footer in custom locations. The horizontal header and footer origins are set to the left edge of the paper here, because our custom headers and footers contain a left tab stop to control placement of the left header and the left footer. The Cocoa printing system's default header and footer contain only center and right tab stops, so if you use them you must draw them at a non-zero point (for example, with an x member set to [printInfo leftMargin]). Remove the following code and call super instead if you want to print the default header and footer returned by -pageHeader and -pageFooter at the default locations set by the Cocoa printing system.
 	if ([[[printInfo dictionary] objectForKey:NSPrintHeaderAndFooter] boolValue]) {
		NSRect imageablePageBounds = [printInfo imageablePageBounds];
		CGFloat spacingFactor = 1.5; // applied to header and footer height for height of gap from content
		CGFloat originX = ([printInfo leftMargin] / scalingFactor) - [printInfo leftMargin]; // scale left edge of paper so left header and footer tab stops coincide with left margin
		
		// Print header.
		CGFloat headerOriginY = ((borderSize.height - [printInfo topMargin]) / scalingFactor) + ([[self pageHeader] size].height * spacingFactor * scalingFactor);
		CGFloat imageablePageTop = (imageablePageBounds.origin.y + imageablePageBounds.size.height) / scalingFactor;
		if (headerOriginY + ([[self pageHeader] size].height / scalingFactor) > imageablePageTop) {
			headerOriginY = imageablePageTop - ([[self pageHeader] size].height / scalingFactor);
		}
		[[self pageHeader] drawAtPoint:NSMakePoint(originX, headerOriginY)];
		
		// Print footer.
		CGFloat footerOriginY = ((borderSize.height - [printInfo topMargin]) / scalingFactor) - borderSize.height + [printInfo topMargin] + [printInfo bottomMargin] - ([[self pageHeader] size].height * spacingFactor) - ([[self pageHeader] size].height * scalingFactor);
		CGFloat imageablePageBottom = imageablePageBounds.origin.y;
		if (footerOriginY < imageablePageBottom) {
			footerOriginY = imageablePageBottom;
		}
		[[self pageFooter] drawAtPoint:NSMakePoint(originX, footerOriginY)];
	}
}

- (NSAttributedString *)pageHeader { // ADDED IN RECIPE 9
	// Provide a custom page header. Called by -drawPageBorderWithSize:. Like the Cocoa printing system's default header, we provide the print job's title in the left header and a date in the right header. The date we provide is either the date printed or the date saved, depending on the print panel's printTimestamp setting, and we label the date accordingly. If the document has not been saved, we replace the date with a string so stating. For horizontal positioning of a left header, a center header and a right header, we set a left tab stop, a center tab stop, and a right tab stop, and we include tab characters in the header text to position individual header elements. Because we include a left tab stop, the header should always be printed against the left edge of the paper (origin.x = 0.0), except when scaling. This differs from the Cocoa printing system's default header, which omits a left tab stop and therefore must be printed at the desired horizontal position on the paper. Our way of doing this moves all header horizontal positioning into this method, and the -drawPageBorderWithSize: method is concerned only with vertical positioning.
	NSPrintInfo *printInfo = [[self currentOperation] printInfo];
	
	// Set right header text based on print panel's timestamp setting (date printed or date saved).
	NSDate *date;
	NSString *dateLabel;
	NSInteger printTimestamp = [[[printInfo dictionary] objectForKey:VRDefaultDiaryPrintTimestampKey] integerValue];
	if (printTimestamp == 0) { // date printed
		date = [self currentDate];
		dateLabel = NSLocalizedString(@"Printed", @"diary document print right header label for date printed");
	} else { // date saved
		date = [[self currentDocument] fileModificationDate];
		dateLabel = NSLocalizedString(@"Saved", @"diary document print right header label for date saved");
	}
	
	// Convert date and time to formatted string.
	NSString *displayDate;
	if (date) {
		if (floor(NSFoundationVersionNumber) <= NSFoundationVersionNumber10_5) {
			NSDateFormatter *formatter = [[NSDateFormatter alloc] init];
			[formatter setDateStyle:NSDateFormatterLongStyle];
			[formatter setTimeStyle:NSDateFormatterShortStyle];
			displayDate = [formatter stringFromDate:date];
			[formatter release];
		} else {
			displayDate = [NSDateFormatter localizedStringFromDate:date dateStyle:NSDateFormatterLongStyle timeStyle:NSDateFormatterShortStyle];
		}
	} else {
		displayDate = NSLocalizedString(@"Document Not Saved", @"diary document print right header date if document not saved");
	}

	// Set header text.
	NSMutableAttributedString *header = [[NSMutableAttributedString alloc] initWithString:[NSString stringWithFormat:@"\t%@\t\t%@: %@", [self printJobTitle], dateLabel, displayDate]];
	
	// Set header paragraph style with three tab stops and set font to system font at default size.
	NSMutableParagraphStyle *paragraphStyle = [[NSParagraphStyle defaultParagraphStyle] mutableCopy];
	[paragraphStyle setTabStops:[self headerAndFooterTabStops]];
	[header setAttributes:[NSDictionary dictionaryWithObjectsAndKeys:paragraphStyle, NSParagraphStyleAttributeName, [NSFont systemFontOfSize:0], NSFontAttributeName, nil] range:NSMakeRange(0, [header length])];
	[paragraphStyle release];
	
	return [header autorelease];
}

- (NSAttributedString *)pageFooter { // ADDED IN RECIPE 9
	// Provide a custom page footer. Called by -drawPageBorderWithSize:. We use the Cocoa printing system's default footer, which displays only the page number and total page count. The only change we make is to provide three tab stops, as described in -pageHeader, and change the font. The default footer contains only two tab stops, so it prints as a right footer when you use the Cocoa text system's default header and footer system. Our three-stop system causes the default footer to be centered because it contains only two tab characters.
	NSMutableAttributedString *footer = [[super pageFooter] mutableCopy];

	// Set footer paragraph style with three tab stops and set font to system font at default size.
	NSMutableParagraphStyle *paragraphStyle = [[NSParagraphStyle defaultParagraphStyle] mutableCopy];
	[paragraphStyle setTabStops:[self headerAndFooterTabStops]];
	[footer setAttributes:[NSDictionary dictionaryWithObjectsAndKeys:paragraphStyle, NSParagraphStyleAttributeName, [NSFont systemFontOfSize:0], NSFontAttributeName, nil] range:NSMakeRange(0, [footer length])];
	[paragraphStyle release];
	
	return [footer autorelease];
}

- (void)drawRect:(NSRect)dirtyRect { // ADDED IN RECIPE 9
	// Called by the Cocoa text system after -drawPageBorderWithSize: and its supporting methods and before -endPage. This method is called for every page. Draw the content of the current page at the point returned by -rectForPage:. Ignore dirtyRect.
	NSLayoutManager *layoutManager = [[[self textStorage] layoutManagers] objectAtIndex:0];
	NSUInteger currentPage = [[self currentOperation] currentPage];
	NSRange glyphRange = [layoutManager glyphRangeForTextContainer:[[layoutManager textContainers] objectAtIndex:currentPage - 1]]; // glyphs were generated and laid out in -paginateContent:
	NSRect pageRect = [self rectForPage:currentPage];
	
	NSPrintInfo *printInfo = [[self currentOperation] printInfo];
	CGFloat scalingFactor;
	if (floor(NSAppKitVersionNumber) <= NSAppKitVersionNumber10_5) {
		scalingFactor = [[[printInfo dictionary] objectForKey:NSPrintScalingFactor] floatValue];
	} else {
		scalingFactor = [printInfo scalingFactor];
	}
	
	[layoutManager drawGlyphsForGlyphRange:glyphRange atPoint:NSMakePoint(pageRect.origin.x, pageRect.origin.y / scalingFactor)]; // glyphs were generated and laid out in -paginateContent:; requires view to be flipped
}

/* Not needed in this implementation.
- (void)endPage { // ADDED IN RECIPE 9
	// Called by the Cocoa text system after after -drawRect: and before -endDocument. This method is called for every page. You must call super at the end of the method because it unlocks focus. We implement this method only for demonstration purposes; if you have no code here, delete this override.	
	// Place any page-specific code here:
	
	[super endPage];
}
*/

/* Not needed in this implementation.
- (void)endDocument { // ADDED IN RECIPE 9
	// Called by the Cocoa text system after -endPage. This method is called for the entire document. You must call super at the end of the method. We implement this method only for demonstration purposes; if you have no code here, delete this override.	
	// Place any document code here.
	
	[super endDocument];
}
*/

#pragma mark UTILITY METHODS

- (NSArray *)headerAndFooterTabStops { // ADDED IN RECIPE 9
	// Return array of three tab stops. Called by -pageHeader and -pageFooter.
	NSPrintInfo *printInfo = [[self currentOperation] printInfo];
	NSTextTab *leftTab = [[NSTextTab alloc] initWithType:NSLeftTabStopType location:[printInfo leftMargin]];
	NSTextTab *centerTab = [[NSTextTab alloc] initWithType:NSCenterTabStopType location:[printInfo paperSize].width / 2];
	NSTextTab *rightTab = [[NSTextTab alloc] initWithType:NSRightTabStopType location:[printInfo paperSize].width - [printInfo rightMargin]];
	return [NSArray arrayWithObjects:leftTab, centerTab, rightTab, nil];
}

- (NSTextStorage *)printableContent { // ADDED IN RECIPE 9
	// Construct the entire printable content of the document by creating a text storage and then adding text from the diary document's diaryDocTextStorage according to the user's printInfo settings. Called from -knowsPageRange.
	NSTextStorage *storage = [[NSTextStorage alloc] init];
	DiaryDocument *document = [self currentDocument];
	DiaryWindowController *windowController = [[document windowControllers] objectAtIndex:0];
	NSDictionary *printInfoDictionary = [[[self currentOperation] printInfo] dictionary];
	
	// Set text storage contents according to printItems setting in print panel.
	NSInteger printItems = [[printInfoDictionary objectForKey:VRDefaultDiaryPrintItemsKey] integerValue];
	switch (printItems) {
		case 0: // entire document contents
			[storage setAttributedString:[document diaryDocTextStorage]];
			break;
		case 1: { // current entry
			NSUInteger insertionPointIndex = [windowController insertionPointIndex];
			NSUInteger startLocation = [document currentEntryTitleRangeForIndex:insertionPointIndex].location;
			NSUInteger endLocation = [document nextEntryTitleRangeForIndex:insertionPointIndex].location;
			if (endLocation == NSNotFound) endLocation = [[document diaryDocTextStorage] length];
			[storage setAttributedString:[[document diaryDocTextStorage] attributedSubstringFromRange:NSMakeRange(startLocation, endLocation - startLocation)]];
			break;
		}
		case 2: { // selection
			NSRange firstSelectionRange = [[[[windowController keyDiaryView] selectedRanges] objectAtIndex:0] rangeValue];
			[storage setAttributedString:[[document diaryDocTextStorage] attributedSubstringFromRange:firstSelectionRange]];
			break;
		}
	}
	
	// Strip tag lists from text storage contents based on printTags setting in print panel. A tag list is identified by paragraph; that is, the tagmarker must be the first character in the paragraph, and the entire paragraph to its end is removed.
	BOOL printTags = [[printInfoDictionary objectForKey:VRDefaultDiaryPrintTagsKey] boolValue];
	if (!printTags) {
#if MAC_OS_X_VERSION_MIN_REQUIRED < MAC_OS_X_VERSION_10_6
		NSMutableArray *removeRangeArray = [NSMutableArray array];
		NSRange tagRange = [DiaryDocument firstTagRangeInString:[storage string]];
		if (tagRange.location != NSNotFound) {
			while (tagRange.location != NSNotFound) {
				[removeRangeArray addObject:[NSValue valueWithRange:tagRange]];
				tagRange = [DiaryDocument nextTagRangeForIndex:NSMaxRange(tagRange) - 1 inString:[storage string]];
			}
			
			for (NSInteger idx = [removeRangeArray count] - 1; idx >= 0; idx--) {
				NSRange substringRange = [[removeRangeArray objectAtIndex:idx] rangeValue];
				[storage deleteCharactersInRange:[[storage string] paragraphRangeForRange:substringRange]];
			}
		}
#else
		[[storage string] enumerateSubstringsInRange:NSMakeRange(0, [storage length]) options:NSStringEnumerationByParagraphs | NSStringEnumerationReverse usingBlock:^(NSString *substring, NSRange substringRange, NSRange enclosingRange, BOOL *stop) {
			if ([substring hasPrefix:[document tagMarker]]) {
				[storage deleteCharactersInRange:[[storage string] paragraphRangeForRange:substringRange]];
			}
		}];
#endif
	}	
	return [storage autorelease];
}
	 
- (void)paginateContent:(NSSize)defaultPageSize { // ADDED IN RECIPE 9
	// Paginate the entire printable content of the document by creating a layout manager and then adding a text container for each page. Reduces individual page size from defaultPageSize as needed to force page breaks to fall on line boundaries. The Cocoa text system will automatically flow the constructed text storage across all text containers for previews and printing. We don't create a text view object for each text container because the text container is used only to calculate pagination. Called from -knowsPageRange.
	NSLayoutManager *layoutManager = [[NSLayoutManager alloc] init];	
	[[self textStorage] addLayoutManager:layoutManager];
	NSTextContainer *textContainer = nil;
	
	NSUInteger glyphCount = [layoutManager numberOfGlyphs];
	NSUInteger lineLocationInGlyphs = 0;
	NSRange lineRangeInGlyphs;
	NSRect lineRect;
	CGFloat pageHeight;
	
	do {
		// Process each line of the text storage separately to accommodate differing line heights based on different fonts, font sizes and styles.		
		if ((lineLocationInGlyphs == 0) || ((pageHeight + lineRect.size.height) > defaultPageSize.height)) {
			// Create a new page at beginning of pagination and again after last line of every subsequent page.
			
			// Adjust final size of text container for previous page, if any.
			if (lineLocationInGlyphs > 0) { // don't execute at beginning of pagination
				[textContainer setContainerSize:NSMakeSize([textContainer containerSize].width, pageHeight)]; // force page break on line boundary
				[textContainer release];
			}
			
			// Create text container for next page at full page size.
			textContainer = [[NSTextContainer alloc] initWithContainerSize:defaultPageSize];
			[layoutManager addTextContainer:textContainer];
			pageHeight = 0.0;
		}
		
		// Lay out text for current line based on local text storage
		if (glyphCount > 0) {
			lineRect = [layoutManager lineFragmentRectForGlyphAtIndex:lineLocationInGlyphs effectiveRange:&lineRangeInGlyphs]; // generates and lays out glyphs
			pageHeight += lineRect.size.height;
			lineLocationInGlyphs = NSMaxRange(lineRangeInGlyphs); // start of next line
		}
	} while (lineLocationInGlyphs < glyphCount);
	
	[textContainer release];
	[layoutManager release];
}

#pragma mark ALERTS

- (NSAlert *)alertCannotPrintScaledUp { // ADDED IN RECIPE 9
	NSBeep();
	NSAlert *alert = [[[NSAlert alloc] init] autorelease];
	[alert setMessageText:NSLocalizedString(@"Cannot print larger than 100%.", @"message text for CannotPrintScaledUp alert")];
	[alert setInformativeText:NSLocalizedString(@"Prints at 100% when Scale is larger.", @"informative text for CannotPrintScaledUp alert")];
	[alert setShowsSuppressionButton:YES];
	return alert;
}

@end

NSString *VRDefaultSuppressScaleUpAlertKey = DEFAULT_ALERT_PRINT_SCALED_UP_SUPPRESSED_KEY; // ADDED IN RECIPE 10
